import argparse
import sys
from operatorcourier import api

def main():
    """Generate the CLI bits
    """
    parser = _CliParser()
    parser.parse()

class _CliParser():
    """Class that generates the command line bits for the operator-courier cli tool
    """

    def __init__(self):
        pass

    def parse(self):
        """Parse generates and evaluates the command level parser
        """
        parser = argparse.ArgumentParser(
            description='Build, verify and push operator bundles into external app registry',
            usage='''operator-courier <command> [<args>]

These are the commands you can use:
    verify      Create a bundle and test it for correctness.
    push        Create a bundle, test it, and push it to an app registry.
    nest        Take a flat to-be-bundled directory and version nest it.
''')
        parser.add_argument('command', help='Subcommand to run')

        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            parser.error('Unrecognized command')
        
        getattr(self, args.command)()

    # Parse the verify command
    def verify(self):
        parser = argparse.ArgumentParser(description='Build and verify an operator bundle to test')
        parser.add_argument('source_dir', help='Path of your directory of yaml files to bundle. Either set this or use the files argument for bundle data.')
        
        args, leftovers = parser.parse_known_args(sys.argv[2:])        
        api.build_and_verify(source_dir=args.source_dir)
    
    # Parse the push command
    def push(self):
        parser = argparse.ArgumentParser(description='Build, verify and push an operator bundle into external app registry.')
        parser.add_argument('source_dir', help='Path of your directory of yaml files to bundle.')
        parser.add_argument('namespace', help='Name of the Quay namespace to push operator to.')
        parser.add_argument('repository', help='Application repository name the application is bundled for.')
        parser.add_argument('release', help='The release version of the bundle.')
        parser.add_argument('token', help='Authorization token for Quay api.')

        args, leftovers = parser.parse_known_args(sys.argv[2:])
        api.build_verify_and_push(args.namespace, args.repository, args.release, args.token, source_dir=args.source_dir)

    # Parse the push command
    def nest(self):
        parser = argparse.ArgumentParser(description='Take a flat bundle directory and version nest it to eventually create an operator-registry image.')
        parser.add_argument('source_dir', help='Path of your directory of yaml files to bundle.')
        parser.add_argument('registry_dir', help='Path of your directory to be populated. If directory does not exist, it will be created.')

        args, leftovers = parser.parse_known_args(sys.argv[2:])
        api.nest(args.source_dir, args.registry_dir)



import yaml
import operatorcourier.identify as identify

class BuildCmd():
    def _get_empty_bundle(self):
         return dict(
            data = dict(
                customResourceDefinitions = [],
                clusterServiceVersions = [],
                packages = [],
            )
        )
    
    def _get_field_entry(self, yamlContent):
        yaml_type = identify.get_operator_artifact_type(yamlContent)
        return yaml_type[0:1].lower() + yaml_type[1:] + 's'

    def _updateBundle(self, operatorBundle, yamlContent):
        operatorBundle["data"][self._get_field_entry(yamlContent)].append(yaml.safe_load(yamlContent))
        return operatorBundle

    def build_bundle(self, strings):
        """build_bundle takes an array of yaml files and generates a 'bundle' with those yaml files
        generated in the bundle format.

        :param strings: Array of yaml strings to bundle
        """
        bundle =  self._get_empty_bundle()
        for item in strings:
            bundle = self._updateBundle(bundle, item)

        return bundle



import yaml

def get_operator_artifact_type(operatorArtifactString):
    """get_operator_artifact_type takes a yaml string and determines if it is one of the 
    expected bundle types: ClusterServiceVersion, CustomResourceDefinition, or Package.

    :param operatorArtifactString: Yaml string to type check
    """
    operatorArtifact = yaml.safe_load(operatorArtifactString)
    if isinstance(operatorArtifact, dict):
        if "packageName" in operatorArtifact:
            return "Package"
        elif "kind" in operatorArtifact:
            if operatorArtifact["kind"] in ("ClusterServiceVersion", "CustomResourceDefinition"):
                return operatorArtifact["kind"]
    raise ValueError('Courier requires valid CSV, CRD, and Package files')



import logging
import json

logger = logging.getLogger(__name__)

class ValidateCmd():
    dataKey = "data"
    crdKey = "customResourceDefinitions"
    csvKey = "clusterServiceVersions"
    pkgsKey = "packages"

    def __init__(self):
        pass
        
    def validate(self, bundle):
        """validate takes a bundle as a dictionary and returns a boolean value that
        describes if the bundle is valid. It also logs verification information when
        there are invalid sections in the bundle.

        :param bundle: Dictionary of bundle value
        """
        logger.info("Validating bundle.")

        return self._bundle_validation(bundle)

    def _bundle_validation(self, bundle):
        validationDict = dict()

        if self.dataKey not in bundle:
            logger.error("Bundle does not contain base data field.")
            #break here because there's nothing else to learn
            return False

        bundleData = bundle[self.dataKey]

        validationDict[self.crdKey] = self._type_validation(bundleData, self.crdKey, self._crd_validation, False)  
        validationDict[self.csvKey] = self._type_validation(bundleData, self.csvKey, self._csv_validation, True)
        validationDict[self.pkgsKey] = self._type_validation(bundleData, self.pkgsKey, self._pkgs_validation, True)

        valid = True
        for key, value in validationDict.items():
            if value is False:
                valid = False
        
        return valid

    def _crd_validation(self, bundleData):
        logger.info("Validating custom resource definitions.")
        valid = True

        crds = bundleData[self.crdKey]

        for crd in crds:
            if "metadata" in crd:
                if "name" in crd["metadata"]:
                    logger.info("Evaluating crd %s", crd["metadata"]["name"])
                else:
                    logger.error("crd metadata.name not defined.")
                    valid = False
            else:
                logger.error("crd metadata not defined.")
                valid = False

            if "apiVersion" not in crd:
                logger.error("crd apiVersion not defined.")
                valid = False

            if "spec" not in crd:
                logger.error("crd spec not defined.")
                valid = False

        return valid

    def _csv_validation(self, bundleData):
        valid = True
        logger.info("Validating cluster service versions.")

        csvs = bundleData[self.csvKey]

        for csv in csvs:
            if "metadata" in csv:
                if self._csv_metadata_validation(csv["metadata"]) is False:
                    valid = False
            else:
                logger.error("csv metadata not defined.")
                valid = False

            if "apiVersion" not in csv:
                logger.error("csv apiVersion not defined.")
                valid = False

            if "spec" in csv:
                if self._csv_spec_validation(csv["spec"], bundleData) is False:
                    valid = False
            else:
                logger.error("csv spec not defined.")
                valid = False

        return valid

    def _csv_spec_validation(self, spec, bundleData):
        valid = True

        warnSpecList = ["displayName", "description", "icon", "version", "provider", "maturity"]
        
        for item in warnSpecList:
            if item not in spec:
                logger.warning("csv spec.%s not defined" % item)

        if "installModes" not in spec:
            logger.error("csv spec.installModes not defined")
            valid = False

        if "install" not in spec:
            logger.error("csv spec.install not defined")
            valid = False

        if "customresourcedefinitions" in spec:
            customresourcedefinitions = spec["customresourcedefinitions"]

            crdList = []
            for crd in bundleData[self.crdKey]:
                try:
                    name = crd["metadata"]["name"]
                    crdList.append(name)
                except KeyError:
                    pass

            if "owned" not in customresourcedefinitions:
                logger.error("spec.customresourcedefinitions.owned not defined for csv")
                valid = False
            else:
                for crd in customresourcedefinitions["owned"]:
                    if "name" not in crd:
                        logger.error("name not defined for item in spec.customresourcedefinitions.")
                        valid = False
                    else:
                        if crd["name"] not in crdList:
                            logger.error("custom resource definition referenced in csv not defined in root list of crds")

        return valid

    def _csv_metadata_validation(self, metadata):
        valid = True

        if "name" in metadata:
            logger.info("Evaluating csv %s", metadata["name"])
        else:
            logger.error("csv metadata.name not defined.")
            valid = False

        if "annotations" in metadata:
            annotations = metadata["annotations"]

            annotationList = ["categories", "description", "containerImage", "createdAt", "support"]

            for item in annotationList:
                if item not in annotations:
                    logger.warning("csv metadata.annotations.%s not defined" % item)

            # check certified value's type in particular. should be string, not bool
            if "certified" not in annotations:
                logger.warning("csv metadata.annotations.certified not defined.")
            else:
                isString = isinstance(annotations["certified"], str)
                if not isString:
                    logger.error("metadata.annotations.certified is not of type string")
                    valid = False

            # if alm-examples is defined, check that its value is valid json
            if "alm-examples" in annotations:
                try:
                    valid_json = json.loads(annotations["alm-examples"])
                except KeyError:
                    logger.error("metadata.annotations.alm-examples contains invalid json string")
                    valid = False

        else:
            logger.warning("csv metadata.annotations not defined.")

        return valid

    def _pkgs_validation(self, bundleData):
        valid = True
        logger.info("Validating packages.")

        pkgs = bundleData[self.pkgsKey]

        for pkg in pkgs:
            if "packageName" in pkg:
                logger.info("Evaluating package %s", pkg["packageName"])
            else:
                logger.error("packageName not defined.")
                valid = False

            if "channels" in pkg:
                channels = pkg["channels"]
                if len(channels) == 0:
                    logger.error("no package channels defined.")
                    valid = False
                else:
                    csvNames = []
                    for csv in bundleData[self.csvKey]:
                        try:
                            csvName = csv["metadata"]["name"]
                            csvNames.append(csvName)
                        except KeyError:
                            pass
                    for channel in channels:
                        if "name" not in channel:
                            logger.error("package channel.name not defined.")
                            valid = False

                        if "currentCSV" not in channel:
                            logger.error("package channel.currentCSV not defined.")
                        else:
                            if channel["currentCSV"] not in csvNames:
                                logger.error("channel.currentCSV %s is not included in list of csvs")
                                valid = False

            else:
                logger.error("package channels not defined.")
                valid = False

        return valid

    def _type_validation(self, bundleData, typeName, validator, required):
        valid = True
        if typeName in bundleData:
            if len(bundleData[typeName]) != 0:
                valid = validator(bundleData)
            elif required:
                logger.error("Bundle does not contain any %s." % typeName)
                valid = False
        elif required:
            logger.error("Bundle does not contain any %s." % typeName)
            valid = False

        return valid



"""
operatorcourier.api module

This module implements the api that should be imported when using the operator-courier package
"""

import os
import logging
from tempfile import TemporaryDirectory
import yaml
from operatorcourier.build import BuildCmd
from operatorcourier.validate import ValidateCmd
from operatorcourier.push import PushCmd
from operatorcourier.format import format_bundle
from operatorcourier.nest import nest_bundles

logging.basicConfig()
logger = logging.getLogger(__name__)

def build_and_verify(source_dir=None, yamls=None):
    """Build and verify constructs an operator bundle from a set of files and then verifies it for usefulness and accuracy.
    It returns the bundle as a string.

    :param source_dir: Path to local directory of yaml files to be read.
    :param yamls: List of yaml strings to create bundle with
    """

    if source_dir is not None and yamls is not None:
        logger.error("Both source_dir and yamls cannot be defined.")
        raise TypeError("Both source_dir and yamls cannot be specified on function call.")

    yaml_files = []

    if source_dir is not None: 
        for filename in os.listdir(source_dir):
            if filename.endswith(".yaml") or filename.endswith(".yml"):
                with open(source_dir + "/" + filename) as f:
                    yaml_files.append(f.read())
    elif yamls is not None:
        yaml_files = yamls

    bundle = BuildCmd().build_bundle(yaml_files)

    valid = ValidateCmd().validate(bundle)

    if not valid:
        bundle = None
        logger.error("Bundle failed validation.")
    else:
        bundle = format_bundle(bundle)
    
    return bundle

def build_verify_and_push(namespace, repository, revision, token, source_dir=None, yamls=None):
    """Build verify and push constructs the operator bundle, verifies it, and pushes it to an external app registry.
    Currently the only supported app registry is the one located at Quay.io (https://quay.io/cnr/api/v1/packages/)

    :param namespace: Quay namespace where the repository we are pushing the bundle is located.
    :param repository: Application repository name the application is bundled for.
    :param revision: Release version of the bundle.
    :param source_dir: Path to local directory of yaml files to be read
    :param yamls: List of yaml strings to create bundle with
    """
    
    bundle = build_and_verify(source_dir, yamls)

    if bundle is not None:
        with TemporaryDirectory() as temp_dir:
            with open('%s/bundle.yaml' % temp_dir, 'w') as outfile:
                yaml.dump(bundle, outfile, default_flow_style=False)
                outfile.flush()

            PushCmd().push(temp_dir, namespace, repository, revision, token)
    else:
        logger.error("Bundle is invalid. Will not attempt to push.")
        raise ValueError("Resulting bundle is invalid, input yaml is improperly defined.")

def nest(source_dir, registry_dir):
    """Nest takes a flat bundle directory and version nests it to eventually be consumed as part of an operator-registry image build.

    :param source_dir: Path to local directory of yaml files to be read
    :param output_dir: Path of your directory to be populated. If directory does not exist, it will be created.
    """
    
    yaml_files = []

    if source_dir is not None: 
        for filename in os.listdir(source_dir):
            if filename.endswith(".yaml") or filename.endswith(".yml"):
                with open(source_dir + "/" + filename) as f:
                    yaml_files.append(f.read())

    with TemporaryDirectory() as temp_dir:
        nest_bundles(yaml_files, registry_dir, temp_dir)



import yaml
from operatorcourier.build import BuildCmd

class _literal(str): pass

def _literal_presenter(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')

def _get_empty_formatted_bundle():
    return dict(
        data = dict(
            customResourceDefinitions = '',
            clusterServiceVersions = '',
            packages = '',
        )
    )

def format_bundle(bundle):
    """
    Converts a bundle object into a push-ready bundle by changing list values of 'customResourceDefinitions', 'clusterServiceVersions', and 'packages' into stringified yaml literals.
    This format is required by the Marketplace backend.
    
    :param bundle: A bundle object
    """

    formattedBundle = _get_empty_formatted_bundle()

    yaml.add_representer(_literal, _literal_presenter)

    if 'data' in bundle:
        # Format data fields as string literals to match backend expected format
        if 'customResourceDefinitions' in bundle['data']:
            if len(bundle['data']['customResourceDefinitions']) > 0:
                formattedBundle['data']['customResourceDefinitions'] = _literal(yaml.dump(bundle['data']['customResourceDefinitions'], default_flow_style=False))

        if 'clusterServiceVersions' in bundle['data']:
            # Format description and alm-examples
            clusterServiceVersions = []
            for csv in bundle['data']['clusterServiceVersions']:
                if 'metadata' in csv:
                    if 'annotations' in csv['metadata']:
                        if 'alm-examples' in csv['metadata']['annotations']:
                            csv['metadata']['annotations']['alm-examples'] = _literal(csv['metadata']['annotations']['alm-examples'])
                
                if 'spec' in csv:
                    if 'description' in csv['spec']:
                        csv['spec']['description'] = _literal(csv['spec']['description'])
                
                clusterServiceVersions.append(csv)
            
            if len(clusterServiceVersions) > 0:
                formattedBundle['data']['clusterServiceVersions'] = _literal(yaml.dump(clusterServiceVersions, default_flow_style=False))

        if 'packages' in bundle['data']:
            if len(bundle['data']['packages']) > 0:
                formattedBundle['data']['packages'] = _literal(yaml.dump(bundle['data']['packages'], default_flow_style=False))

    return formattedBundle

def unformat_bundle(formattedBundle):
    """
    Converts a push-ready bundle into a structured object by changing stringified yaml of 'customResourceDefinitions', 'clusterServiceVersions', and 'packages' into lists of objects.
    Undoing the format helps simplify bundle validation.

    :param formattedBundle: A push-ready bundle
    """

    bundle = BuildCmd()._get_empty_bundle()
 
    if 'data' in formattedBundle:
        if 'customResourceDefinitions' in formattedBundle['data']:
            customResourceDefinitions = yaml.safe_load(formattedBundle['data']['customResourceDefinitions'])
            if customResourceDefinitions and len(customResourceDefinitions) > 0:
                bundle['data']['customResourceDefinitions'] = customResourceDefinitions
        
        if 'clusterServiceVersions' in formattedBundle['data']:
            clusterServiceVersions = yaml.safe_load(formattedBundle['data']['clusterServiceVersions'])
            if clusterServiceVersions and len(clusterServiceVersions) > 0:
                bundle['data']['clusterServiceVersions'] = clusterServiceVersions

        if 'packages' in formattedBundle['data']:
            packages = yaml.safe_load(formattedBundle['data']['packages'])
            if packages and len(packages) > 0:
                bundle['data']['packages'] = packages

    return bundle



from setuptools import setup
setup(
  name='operator-courier',
  packages=['operatorcourier'],
  version='1.0.2',
  description='Build, verify and push operators',
  author='AOS Marketplace',
  author_email='aos-marketplace@redhat.com',
  url='https://github.com/operator-framework/operator-courier', 
  entry_points = {
    'console_scripts': ['operator-courier=operatorcourier.cli:main'],
  },
  keywords = ['operator', 'courier'],
  install_requires=[
    'pyyaml', 
    'requests'
  ],
  setup_requires=['pytest-runner'],
  tests_require=['pytest'],
)



from distutils.dir_util import copy_tree
import logging
import os
import yaml
import operatorcourier.identify as identify

logger = logging.getLogger(__name__)

def nest_bundles(yaml_files, registry_dir, temp_registry_dir):
    package = {}
    crds = {}
    csvs = []

    errors = []

    # first lets parse all the files
    for yaml_string in yaml_files:
        yaml_type = identify.get_operator_artifact_type(yaml_string)
        if yaml_type == "Package":
            if not package:
                package = yaml.safe_load(yaml_string)
            else:
                errors.append("Multiple packages in directory.")
        if yaml_type == "CustomResourceDefinition":
            crd = yaml.safe_load(yaml_string)
            crd_name = crd["metadata"]["name"]
            crds[crd_name] = crd
        if yaml_type == "ClusterServiceVersion":
            csv = yaml.safe_load(yaml_string)
            csvs.append(csv)

    if len(csvs) == 0:
        errors.append("No csvs in directory.")

    if not package:
        errors.append("No package file in directory.")

    # write the package file
    package_name = package["packageName"]
    with open('%s/%s.package.yaml' % (temp_registry_dir, package_name), 'w') as outfile:
        yaml.dump(package, outfile, default_flow_style=False)
        outfile.flush()

    # now lets create a subdirectory for each version of the csv, and add all the relevant crds to it
    for csv in csvs:
        csv_name = csv["metadata"]["name"]
        version = csv["spec"]["version"]
        csv_folder = temp_registry_dir + "/" + version

        if not os.path.exists(csv_folder):
            os.makedirs(csv_folder)

        with open('%s/%s.clusterserviceversion.yaml' % (csv_folder, csv_name), 'w') as outfile:
            yaml.dump(csv, outfile, default_flow_style=False)
            outfile.flush()

        csv_crds = csv["spec"]["customresourcedefinitions"]["owned"]
        for csv_crd in csv_crds:
            crd_name = csv_crd["name"]
            if crd_name in crds:
                crd = crds[crd_name]
                with open('%s/%s.crd.yaml' % (csv_folder, crd_name), 'w') as outfile:
                    yaml.dump(crd, outfile, default_flow_style=False)
                    outfile.flush()
            else:
                errors.append("CRD %s mentioned in CSV %s was not found in directory." % (crd_name, csv_name))

    # if no errors were encountered, lets create the real directory and populate it.
    if len(errors) == 0:
        if not os.path.exists(registry_dir):
            os.makedirs(registry_dir)
        copy_tree(temp_registry_dir, registry_dir)
    else:
        for err in errors:
            logger.error(err)
